@@ -2,7 +2,8 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017-2020 Osimis S.A., Belgium
+ * Copyright (C) 2017-2023 Osimis S.A., Belgium
+ * Copyright (C) 2021-2023 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Affero General Public License
@@ -21,7 +22,6 @@
 
 #include "BinaryStringValue.h"
 
-#include "FileValue.h"
 #include "NullValue.h"
 
 #include <OrthancException.h>
@@ -34,9 +34,6 @@ namespace OrthancDatabases
   {
     switch (target)
     {
-      case ValueType_File:
-        return new FileValue(content_);
-
       case ValueType_Null:
         return new NullValue;
 
@@ -44,10 +41,4 @@ namespace OrthancDatabases
         throw Orthanc::OrthancException(Orthanc::ErrorCode_BadParameterType);
     }
   }
-
-
-  std::string BinaryStringValue::Format() const
-  {
-    return "(binary - " + boost::lexical_cast<std::string>(content_.size()) + " bytes)";
-  }
 }
